package net.beaumaris.cachetest;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import org.infinispan.Cache;
import org.infinispan.distexec.DefaultExecutorService;
import org.infinispan.distexec.DistributedExecutorService;
import org.infinispan.manager.DefaultCacheManager;
import org.infinispan.transaction.tm.DummyTransactionManager;

public class BaseCacheTest {

   protected final static int NUM_NODES = 3;
   protected final static int NUM_THREADS = 2;
   protected final static int COUNT = 5000;

   protected final static int KEY_UPPER_BOUND = 100;
   protected final static int INCREMENT_AMOUNT = 1000;

   protected static Random r = new Random();

   protected List<DefaultCacheManager> cacheContainers = Collections.synchronizedList(new ArrayList<DefaultCacheManager>(NUM_NODES));
   protected DummyTransactionManager tm = new DummyTransactionManager();

   public void createNodes() throws InterruptedException, IOException {

      for (int nodeNum = 0; nodeNum < NUM_NODES; nodeNum++) {
         addNode();
      }
   }

   public void shutdown() {

      for (DefaultCacheManager cm : cacheContainers) {

         if (cm != null)
            cm.stop();
      }

      cacheContainers.clear();
   }

   protected void addNode() throws IOException {

      DefaultCacheManager cm = new DefaultCacheManager("cluster.xml");
      cm.start();
      cacheContainers.add(cm);
   }

   public class StressingThread implements Runnable {

      protected static final long SLEEP_MILLISECONDS = 10;

      protected final SharedStats stats;
      protected final CountDownLatch latch;

      protected DistributedExecutorService executor;

      public StressingThread(SharedStats stats, DefaultCacheManager cacheManager, final CountDownLatch latch) {

         this.stats = stats;
         this.latch = latch;

         Cache<Integer, Integer> allocCache = cacheManager.getCache("alloc");
         executor = new DefaultExecutorService(allocCache);
      }

      public void run() {

         int i = 0;
         int errorCount = 0;

         while (i < COUNT && !(stats.globalQuit || Thread.interrupted())) {

            try {
               doCycle();
            } catch (Exception e) {

               errorCount++;
               checkIsTrue(false, e.getMessage());
            }

            i++;
         }

         latch.countDown();
      }

      private void doCycle() throws Exception {

         // Pick a key.
         int key = r.nextInt(KEY_UPPER_BOUND) + 1;

         AddTask addTask = new AddTask(key, INCREMENT_AMOUNT);
         Future<String> addResult = executor.submit(addTask, key);

         stats.allocateSuccessCounter.getAndIncrement();

         String uuid = addResult.get(2000, TimeUnit.MILLISECONDS);

         sleep(1000);

         SubtractTask subtractTask = new SubtractTask(uuid, INCREMENT_AMOUNT);

         Future<Integer> subtractResult = executor.submit(subtractTask, key);
         Integer result = subtractResult.get(2000, TimeUnit.MILLISECONDS);

         if (result == 0)
            stats.deallocateSuccessCounter.getAndIncrement();
         else
            stats.deallocateFailedCounter.getAndIncrement();

         // Wait before retrying.
         sleep(1000);
      }

      private void sleep(long milliseconds) {
         try {
            Thread.sleep(milliseconds);
         } catch (InterruptedException e) {
            // no-op: waking up is good enough
         }
      }

      private void checkIsTrue(boolean assertion, String message) {
         if (assertion == false) {
            stats.seenFailures = true;
            System.out.println(message);
         }
      }
   }

   /**
    * Common state to verify cache behavior
    */
   public static class SharedStats {

      final AtomicInteger allocateSuccessCounter = new AtomicInteger(0);
      final AtomicInteger deallocateSuccessCounter = new AtomicInteger(0);
      final AtomicInteger rollbackCounter = new AtomicInteger(0);
      final AtomicInteger allocateFailedCounter = new AtomicInteger(0);
      final AtomicInteger timeoutCounter = new AtomicInteger(0);
      final AtomicInteger deallocateFailedCounter = new AtomicInteger(0);
      Throwable throwable = null;
      volatile boolean globalQuit = false; // when it's true the threads quit
      volatile boolean seenFailures = false; // set to true by a thread if it
                                             // has experienced

      // illegal state

      public String toString() {
         return "\n\tSuccessful allocation count:\t" + allocateSuccessCounter.get()
                + "\n\tSuccessful deallocation count:\t"
                + deallocateSuccessCounter.get()
                + "\n\tRollback tx count:\t"
                + rollbackCounter.get()
                + "\n\tFailed allocation count:\t"
                + allocateFailedCounter.get()
                + "\n\tFailed deallocation count:\t"
                + deallocateFailedCounter.get()
                + "\n\tOther cache exception count:\t"
                + timeoutCounter.get()
                + "\n\tIllegal state detected:\t"
                + seenFailures;
      }
   }
}
